@@ -28,12 +28,18 @@
 #include <wchar.h>
 
 #include "error.h"
+#include "common.h"
 #include "packet.h"
 #include "sound.h"
 
+extern unsigned char saved_level;
+extern unsigned char level;
+
+extern const char* gameid;
+
 #ifndef WIN32
 typedef int SOCKET;
-const int INVALID_SOCKET = -1;
+#define INVALID_SOCKET (-1)
 const int SOCKET_ERROR = -1;
 
 #define Sleep(n) usleep((n) * 1000)
@@ -49,33 +55,26 @@ static struct option long_options[] =
   { 0, 0, 0, 0 }
 };
 
+extern void play_single_player();
+
 #ifndef swprintf
 int swprintf(wchar_t *wcs, size_t maxlen, const wchar_t *format, ...);
 #endif
 #endif
 
-int showurl = 0;
 int hwalpha;
 
-void load_images();
-SDL_Surface* get_image(const char* name);
-
-void load_font();
-int has_char(int font, int ch);
-void print_string(int font, int x, int y, const wchar_t* string, int align);
-int string_width(int font, const wchar_t* string, size_t length);
-
-const int width = 640;
-const int height = 480;
-int fullscreen = 1;
+int fullscreen = 1; /* XXX: Change to 1 before release */
 int sound_enable = 1;
 SDL_Surface* screen;
 
+int time_stepms = 8;
+float time_step = 0.008f;
 unsigned short listenport;
-SOCKET listenfd;
-SOCKET serverfd;
-SOCKET outpeerfd;
-SOCKET inpeerfd;
+SOCKET listenfd = INVALID_SOCKET;
+SOCKET serverfd = INVALID_SOCKET;
+SOCKET outpeerfd = INVALID_SOCKET;
+SOCKET inpeerfd = INVALID_SOCKET;
 int outpeerconnected;
 int outpeervalid;
 int inpeervalid;
@@ -95,7 +94,7 @@ static int outbufferop_state;
 static int outbuffer_state;
 
 unsigned int tickidx;
-unsigned int last_evil = 0;
+static unsigned int last_evil;
 
 Uint32 last_tick = 0;
 
@@ -114,16 +113,10 @@ SDL_Surface* background;
 SDL_Surface* base[129];
 SDL_Surface* bubbles[8];
 
-const int time_stepms = 8;
-const float time_step = 0.008f;
-const int field_height = 14;
-const int max_field_width = 8;
-const float bubble_speed = 500;
-const int max_x = 414 - 190;
-#define WIDTH(y) (((y) & 1) ? 7 : 8)
-
 enum mode
 {
+  MODE_MODE_SELECT,
+  MODE_SINGLEPLAYER_MENU,
   MODE_MAIN_MENU,
   MODE_LOUNGE,
   MODE_WAITING,
@@ -133,10 +126,11 @@ enum mode
   MODE_ABORT_MESSAGE2
 };
 
-enum mode mode = MODE_MAIN_MENU;
+enum mode mode = MODE_MODE_SELECT;
 wchar_t username[17];
 wchar_t password[17];
-wchar_t message[257];
+wchar_t message[258];
+size_t  message_cursor;  // cursor is before message[message_cursor]
 int messageidx;
 int won;
 int is_server;
@@ -166,56 +160,13 @@ enum flag
   FLAG_READY = 0x20
 };
 
-struct moving_bubble
-{
-  float x;
-  float y;
-  float velx;
-  float vely;
-
-  int lastpaintx;
-  int lastpainty;
-  int evil;
-
-  unsigned char falling;
-  unsigned char color;
-};
-
-struct player_state
-{
-  unsigned char field[20][8];
-
-  int bubble;
-  int next_bubble;
-
-  float angle;
-  int right;
-
-  int dirty_minx;
-  int dirty_miny;
-  int dirty_maxx;
-  int dirty_maxy;
-  float last_angle;
-
-  int score;
-  int ready;
-
-  struct moving_bubble mbubbles[128];
-
-  unsigned char evil_bubbles[128];
-  int evil_bubble_count;
-  unsigned int evil_bubble_seed;
-};
-
 static struct player_state players[2];
 
 static void init_field();
 
-#define SET_RECT(r, xv, yv, wv, hv) do { (r).x = (xv); (r).y = (yv); (r).w = (wv); (r).h = (hv); } while(0)
+unsigned int rng_seed = 0;
 
-static unsigned int rng_seed = 0;
-
-static unsigned int rng()
+unsigned int rng()
 {
   rng_seed = (rng_seed + 0xA853753) % 177
            + (rng_seed * 0x8034052);
@@ -252,7 +203,78 @@ static float unpack_float(const unsigned char* a)
   return u.fval;
 }
 
-static void mark_dirty(struct player_state* p, int x, int y, int width, int height)
+// Calculate length of current message
+// Set cursor at end of msg if too far
+size_t message_length()
+{
+  size_t length = 0;
+  while(length < 256 && message[length])
+    ++length;
+  if (message_cursor > length)
+    message_cursor = length;
+  return length;
+}
+
+// Remove char left of cursor, shift end of message (including trailing \0)
+void message_backspace()
+{
+  size_t length = message_length();
+  if (message_cursor > 0)
+  {
+    memmove(message+message_cursor-1,
+            message+message_cursor,
+            sizeof(message[0])*(length-message_cursor+1));
+    --message_cursor;
+  }
+}
+
+// Remove char right of cursor, shift end of message (including \0) :
+void message_delete()
+{
+  size_t length = message_length();
+  if (message_cursor < length)
+  {
+    memmove(message+message_cursor,
+            message+message_cursor+1,
+            sizeof(message[0])*(length-message_cursor));
+  }
+}
+
+
+// Insert char at current position : must shift end of msg
+void message_insert(wchar_t c)
+{
+  size_t length = message_length();
+  if (length < 256)
+  {
+    memmove(message+message_cursor+1,
+            message+message_cursor,
+            sizeof(message[0])*(length-message_cursor+1));
+    message[message_cursor] = c;
+    ++message_cursor;
+  }
+}
+
+// Try to move cursor one char left
+void cursor_left()
+{
+  if (message_cursor > 0)
+  {
+    --message_cursor;
+  }
+}
+
+// Try to move cursor one char right
+void cursor_right()
+{
+  size_t length = message_length();
+  if (message_cursor < length)
+  {
+    ++message_cursor;
+  }
+}
+
+void mark_dirty(struct player_state* p, int x, int y, int width, int height)
 {
   if(x < p->dirty_minx)
     p->dirty_minx = x;
@@ -274,9 +296,17 @@ static void scan(struct player_state* p, int x, int y, int color, int* matches,
   if(x < 0 || x >= WIDTH(y) || y < 0 || y >= field_height)
     return;
 
-  if((color && (p->field[y][x] != color))
-  || (!color && !p->field[y][x]))
-    return;
+  if(!color)
+  {
+    if(!p->field[y][x])
+      return;
+  }
+  else
+  {
+    if(p->field[y][x] != 9
+    && p->field[y][x] != color)
+      return;
+  }
 
   idx = y * max_field_width + x;
 
@@ -305,7 +335,7 @@ static void scan(struct player_state* p, int x, int y, int color, int* matches,
   }
 }
 
-static void remove_bubble(struct player_state* p, int x, int y, int evil)
+void remove_bubble(struct player_state* p, int x, int y, int evil)
 {
   int i;
   struct moving_bubble mbubble;
@@ -340,6 +370,18 @@ static void chatlog_append(int is_private, const wchar_t* m)
   while(length < 256 && m[length])
     ++length;
 
+  // Check for call for fight, or message containing username :
+  // The username is not searched at the beginning, for we are
+  // not interested in detecting our own words.
+  if (wcsncmp(m, L"*** You got a call for fight", 28) == 0)
+  {
+    sounds[0].pos = 0;  // explosion
+  }
+  if (length > 0 && username[0]!=0 && wcsstr(m+1, username) != NULL)
+  {
+    sounds[2].pos = 0;  // piou!
+  }
+
   if(string_width(1, m, length) > CHAT_WIDTH)
   {
     wchar_t rbuf[267];
@@ -535,26 +577,17 @@ static void init_movement_packet(struct data_packet* p)
   p->movement.bubbles = players[0].bubble | (players[0].next_bubble << 4);
 }
 
-static void stick(struct player_state* p, int bx, int by, int color)
+static int test_group(struct player_state* p, int bx, int by, int color)
 {
-  int i, j, k;
   int matches[max_field_width * field_height];
   int match_count = 0;
-
-  if(by >= field_height)
-    return;
-
-  if(p == &players[0] && sound_enable)
-    sounds[3].pos = 0;
-
-  p->field[by][bx] = color;
-  mark_dirty(p, bx * 32 + ((by & 1) ? 16 : 0), by * 28, 32, 32);
+  int i;
 
   scan(p, bx, by, color, matches, &match_count);
 
   if(match_count >= 3)
   {
-    if(p == &players[0] && sound_enable)
+    if(p != &players[1] && sound_enable)
       sounds[0].pos = 0;
 
     for(i = 0; i < match_count; ++i)
@@ -563,6 +596,52 @@ static void stick(struct player_state* p, int bx, int by, int color)
                     matches[i] / max_field_width, 0);
     }
 
+    return 1;
+  }
+
+  return 0;
+}
+
+int stick(struct player_state* p, int bx, int by, int color)
+{
+  int matches[max_field_width * field_height];
+  int match_count = 0;
+  int i, j, k;
+
+  if(by >= field_height)
+    return 0;
+
+  if(p != &players[1] && sound_enable)
+    sounds[3].pos = 0;
+
+  p->field[by][bx] = color;
+  mark_dirty(p, bx * 32 + ((by & 1) ? 16 : 0), by * 28, 32, 32);
+
+  int broken = 0;
+
+  if(color <= 8)
+  {
+    if(test_group(p, bx, by, color))
+      broken = 1;
+  }
+  else
+  {
+    for(i = 1; i <= 8; ++i)
+    {
+      p->field[by][bx] = i;
+
+      if(test_group(p, bx, by, i))
+        broken = 1;
+    }
+
+    if(broken)
+      p->field[by][bx] = 0;
+    else
+      p->field[by][bx] = color;
+  }
+
+  if(broken)
+  {
     match_count = 0;
 
     for(i = 0; i < WIDTH(0); ++i)
@@ -588,53 +667,65 @@ static void stick(struct player_state* p, int bx, int by, int color)
         }
       }
 
-      for(i = 0; i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++i)
-        p->mbubbles[i].falling = 1;
+      return 1;
     }
-    else
+
+    for(i = 0; i < field_height; ++i)
     {
-      for(i = 0; i < field_height; ++i)
+      for(j = 0; j < WIDTH(i); ++j)
       {
-        for(j = 0; j < WIDTH(i); ++j)
+        if(!p->field[i][j])
+          continue;
+
+        for(k = 0; k < match_count; ++k)
+          if(matches[k] == i * max_field_width + j)
+            break;
+
+        if(k == match_count)
         {
-          if(!p->field[i][j])
-            continue;
-
-          for(k = 0; k < match_count; ++k)
-            if(matches[k] == i * max_field_width + j)
-              break;
-
-          if(k == match_count)
+          if(peer_version < 3)
           {
-            if(peer_version < 3)
+            if(p == &players[0])
             {
-              if(p == &players[0])
-              {
-                k = players[1].evil_bubble_count++;
+              k = players[1].evil_bubble_count++;
 
-                players[1].evil_bubbles[k] = p->field[i][j];
-              }
-              else
-              {
-                k = players[0].evil_bubble_count++;
-
-                players[0].evil_bubbles[k] = p->field[i][j];
-              }
+              players[1].evil_bubbles[k] = p->field[i][j];
             }
+            else
+            {
+              k = players[0].evil_bubble_count++;
 
-            remove_bubble(p, j, i, 1);
+              players[0].evil_bubbles[k] = p->field[i][j];
+            }
           }
+          else
+          {
+            if(p == &players[0])
+            {
+              if(players[1].max_evil < 4)
+                ++players[1].max_evil;
+            }
+            else
+            {
+              if(players[0].max_evil < 4)
+                ++players[0].max_evil;
+            }
+          }
+
+          remove_bubble(p, j, i, 1);
         }
       }
     }
   }
+
+  return 0;
 }
 
-static void cond_blit(struct player_state* p, SDL_Surface* source, SDL_Rect* source_rect, SDL_Surface* dest, SDL_Rect* dest_rect)
+void cond_blit(struct player_state* p, SDL_Surface* source, SDL_Rect* source_rect, SDL_Surface* dest, SDL_Rect* dest_rect)
 {
   int xoff, yoff;
 
-  if(p == &players[0])
+  if(p != &players[1])
     xoff = 30, yoff = 40;
   else
     xoff = 355, yoff = 40;
@@ -828,15 +919,17 @@ collide:;
         {
           if(b->evil && peer_version >= 3)
           {
-            if(p == &players[0])
+            if(p == &players[0] && players[1].max_evil)
             {
               k = players[1].evil_bubble_count++;
+              --players[1].max_evil;
 
               players[1].evil_bubbles[k] = b->color;
             }
-            else
+            else if(p == &players[1] && players[0].max_evil)
             {
               k = players[0].evil_bubble_count++;
+              --players[0].max_evil;
 
               players[0].evil_bubbles[k] = b->color;
             }
@@ -977,11 +1070,14 @@ collide:;
 static int random_bubble(struct player_state* p)
 {
   int i, j, result;
+  int tried = 0;
 
-  for(;;)
+  while(tried != 0xFF)
   {
     result = rand() % 8;
 
+    tried |= (1 << result);
+
     for(i = 0; i < field_height; ++i)
     {
       for(j = 0; j < WIDTH(i); ++j)
@@ -991,17 +1087,20 @@ static int random_bubble(struct player_state* p)
       }
     }
   }
+
+  return rand() % 8;
 }
 
-static void shoot(struct player_state* p)
+void shoot(struct player_state* p, int color, int velocity)
 {
   int i;
   struct moving_bubble mbubble;
+  float speed = (velocity == -1) ? bubble_speed : velocity;
 
   mbubble.falling = 0;
   mbubble.evil = 0;
-  mbubble.velx = sin(p->angle / 180.0f * M_PI) * bubble_speed;
-  mbubble.vely = -cos(p->angle / 180.0f * M_PI) * bubble_speed;
+  mbubble.velx = sin(p->angle / 180.0f * M_PI) * speed;
+  mbubble.vely = -cos(p->angle / 180.0f * M_PI) * speed;
   mbubble.x = 112.0f;
   mbubble.y = 350.0f;
   mbubble.color = p->bubble + 1;
@@ -1021,13 +1120,18 @@ static void shoot(struct player_state* p)
   p->bubble = p->next_bubble;
 
   if(p != &players[1])
-    p->next_bubble = random_bubble(p);
+  {
+    if(color == -1)
+      p->next_bubble = random_bubble(p);
+    else
+      p->next_bubble = color;
+  }
 
   if(sound_enable)
     sounds[1].pos = 0;
 }
 
-static void init_player(struct player_state* p)
+void init_player(struct player_state* p)
 {
   int i;
 
@@ -1036,6 +1140,7 @@ static void init_player(struct player_state* p)
   p->angle = 0.0f;
   p->right = 0;
   p->evil_bubble_count = 0;
+  p->max_evil = 0;
   p->evil_bubble_seed = 0;
 
   for(i = 0; i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]); ++i)
@@ -1093,15 +1198,15 @@ void connect_to_master()
   outbuffer_state = 0;
   inbuffer_state = 0;
 
-  host = gethostbyname("junoplay.com");
+  host = gethostbyname("master.junoplay.com");
 
   if(!host)
     return;
 
-  if(INVALID_SOCKET == (serverfd = socket(PF_INET, SOCK_STREAM, 0)))
+  if(INVALID_SOCKET == (serverfd = socket(host->h_addrtype, SOCK_STREAM, 0)))
     return;
 
-  address.sin_family = AF_INET;
+  address.sin_family = host->h_addrtype;
   address.sin_port = htons(tcp_port);
   memcpy(&address.sin_addr, host->h_addr, host->h_length);
 
@@ -1116,7 +1221,8 @@ void connect_to_master()
   packet.packet_type = packet_identify;
   packet.payload_size = sizeof(packet.identify);
 
-  strcpy(packet.identify.game, "pengupop2006102201");
+  strcpy(packet.identify.game, "pengupop2007011500");
+  memcpy(packet.identify.id, gameid, 32);
 
   if(listenfd == INVALID_SOCKET)
   {
@@ -1147,7 +1253,8 @@ void connect_to_master()
       packet.payload_size = sizeof(packet.login);
 
       if(16 == read(fd, packet.login.username, 16)
-      && 16 == read(fd, packet.login.password, 16))
+      && 16 == read(fd, packet.login.password, 16)
+      && packet.login.username[0])
       {
         int i;
 
@@ -1170,20 +1277,20 @@ void connect_to_master()
   if(auth_level != 3)
   {
     int i;
-    char auth[64];
-    DWORD authsize = sizeof(auth);
+    char buf[64];
+    DWORD bufsize = sizeof(buf);
 
     HKEY k_config;
 
     if(ERROR_SUCCESS == RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Junoplay.com\\Pengupop\\Config", 0, KEY_READ, &k_config))
     {
-      if(ERROR_SUCCESS == RegQueryValueEx(k_config, "auth", 0, 0, auth, &authsize))
+      if(ERROR_SUCCESS == RegQueryValueEx(k_config, "auth", 0, 0, buf, &bufsize))
       {
-        auth[63] = 0;
+        buf[63] = 0;
 
         packet.packet_type = packet_login;
         packet.payload_size = sizeof(packet.login);
-        sscanf(auth, "%[^:]:%[^\n]", packet.login.username, packet.login.password);
+        sscanf(buf, "%[^:]:%s", packet.login.username, packet.login.password);
 
         send_packet(&packet);
 
@@ -1488,7 +1595,7 @@ void process_packet(struct data_packet* packet)
         strcpy(confpath, getenv("HOME"));
         strcat(confpath, "/.pengupoprc");
 
-        int fd = open(confpath, O_WRONLY | O_CREAT | O_TRUNC, 0600);
+        int fd = open(confpath, O_WRONLY | O_CREAT, 0600);
 
         if(fd != -1)
         {
@@ -1557,7 +1664,7 @@ void show_splash()
 
   last_tick = SDL_GetTicks();
 
-  while(SDL_GetTicks() - last_tick < 5000)
+  while(SDL_GetTicks() - last_tick < 3000)
   {
     while(SDL_PollEvent(&event))
     {
@@ -1635,38 +1742,6 @@ static void join_random(int registered_only)
   mode = MODE_GAME;
 }
 
-void toggle_fullscreen()
-{
-#ifndef WIN32
-  SDL_WM_ToggleFullScreen(screen);
-#else
-  if(fullscreen)
-    screen = SDL_SetVideoMode(width, height, 0, SDL_SWSURFACE);
-  else
-    screen = SDL_SetVideoMode(width, height, 0, SDL_SWSURFACE | SDL_FULLSCREEN);
-
-  if(mode != MODE_GAME)
-  {
-    SDL_BlitSurface(logo, 0, screen, 0);
-  }
-  else
-  {
-    int i;
-
-    SDL_BlitSurface(background, 0, screen, 0);
-
-    for(i = 0; i < 2; ++i)
-    {
-      players[i].dirty_minx = 0;
-      players[i].dirty_miny = 0;
-      players[i].dirty_maxx = max_field_width * 32;
-      players[i].dirty_maxy = 440;
-    }
-  }
-#endif
-  fullscreen = !fullscreen;
-}
-
 #ifndef WIN32
 int main(int argc, char** argv)
 #else
@@ -1679,7 +1754,6 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
   SDL_Rect rect;
 
 #ifdef WIN32
-  showurl = 1;
   FreeConsole();
 
   do /* Register `pengupop' URL handler */
@@ -1752,12 +1826,13 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
     case 'v':
 
       printf("%s\n", PACKAGE_STRING);
-      printf("Copyright (C) 2006 Junoplay\n"
-             "This is free software.  You may redistribute copies of it under the terms of\n"
-             "the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.\n"
-             "There is NO WARRANTY, to the extent permitted by law.\n"
-             "\n"
-             "Written by Morten Hustveit.  Artwork by Jorgen Jacobsen.\n");
+      printf(
+        "Copyright (C) 2006 Junoplay\n"
+        "This is free software.  You may redistribute copies of it under the terms of\n"
+        "the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.\n"
+        "There is NO WARRANTY, to the extent permitted by law.\n"
+        "\n"
+        "Written by Morten Hustveit.  Artwork by Jorgen Jacobsen.\n");
 
       return EXIT_SUCCESS;
 
@@ -1806,6 +1881,51 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
   background = get_image("backgrnd.png");
   SDL_SetAlpha(background, 0, 0);
 
+#if LINUX || DARWIN
+  if(getenv("HOME"))
+  {
+    char confpath[4096];
+
+    strcpy(confpath, getenv("HOME"));
+    strcat(confpath, "/.pengupoprc");
+
+    int fd = open(confpath, O_RDONLY);
+
+    if(fd != -1)
+    {
+      lseek(fd, 32, SEEK_SET);
+
+      read(fd, &level, 1);
+
+      if(level >= 201)
+        level = 0;
+
+      close(fd);
+    }
+  }
+#elif defined(WIN32)
+  {
+    char buf[64];
+    DWORD bufsize = sizeof(buf);
+    HKEY k_config;
+
+    if(ERROR_SUCCESS == RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Junoplay.com\\Pengupop\\Config", 0, KEY_READ, &k_config))
+    {
+      if(ERROR_SUCCESS == RegQueryValueEx(k_config, "bananas", 0, 0, buf, &bufsize))
+      {
+        buf[63] = 0;
+
+        int tmp;
+        sscanf(buf, "%d", &tmp);
+        level = tmp ^ 0x7236143;
+
+        if(level >= 201)
+          level = 0;
+      }
+    }
+  }
+#endif
+
   for(i = 0; i < 129; ++i)
   {
     char name[64];
@@ -1829,11 +1949,11 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
   start_listening();
 
-  connect_to_master();
-
 #ifdef WIN32
   if(strlen(cmdline) > 6)
   {
+    connect_to_master();
+
     if(serverfd == INVALID_SOCKET)
     {
       wcscpy(message, L"Failed to Connect to Master Server");
@@ -2300,6 +2420,42 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
         print_string(0, 320, 290, L"(Press Escape)", 1);
       }
     }
+    else if(mode == MODE_MODE_SELECT)
+    {
+      int bump = (int) fabsf(sinf(last_tick * 0.2f) * 10.0f) - 5;
+      ++last_tick;
+
+      Sleep(15);
+
+      rect.x = 0;
+      rect.y = 200;
+      rect.w = 640;
+      rect.h = 260;
+
+      SDL_BlitSurface(logo, &rect, screen, &rect);
+
+      print_string(0, 320 + (selection == 0 ? bump : 0), 250, L"Single Player", 1);
+      print_string(0, 320 + (selection == 1 ? bump : 0), 290, L"Online Multiplayer", 1);
+      print_string(0, 320 + (selection == 2 ? bump : 0), 330, L"Quit", 1);
+    }
+    else if(mode == MODE_SINGLEPLAYER_MENU)
+    {
+      int bump = (int) fabsf(sinf(last_tick * 0.2f) * 10.0f) - 5;
+      ++last_tick;
+
+      Sleep(15);
+
+      rect.x = 0;
+      rect.y = 200;
+      rect.w = 640;
+      rect.h = 260;
+
+      SDL_BlitSurface(logo, &rect, screen, &rect);
+
+      print_string(0, 320 + (selection == 0 ? bump : 0), 250, L"Adventure Mode", 1);
+      print_string(0, 320 + (selection == 1 ? bump : 0), 290, L"Infinite Mode", 1);
+      print_string(0, 320 + (selection == 2 ? bump : 0), 330, L"Quit", 1);
+    }
     else if(mode == MODE_MAIN_MENU)
     {
       int bump = (int) fabsf(sinf(last_tick * 0.2f) * 10.0f) - 5;
@@ -2316,13 +2472,7 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
       print_string(0, 320 + (selection == 0 ? bump : 0), 250, L"Join Random Game", 1);
       print_string(0, 320 + (selection == 1 ? bump : 0), 290, L"Enter Lounge", 1);
-      if(showurl)
-      {
-        print_string(0, 320 + (selection == 2 ? bump : 0), 330, L"Buy Full Version", 1);
-        print_string(0, 320 + (selection == 3 ? bump : 0), 370, L"Quit", 1);
-      }
-      else
-        print_string(0, 320 + (selection == 2 ? bump : 0), 330, L"Quit", 1);
+      print_string(0, 320 + (selection == 2 ? bump : 0), 330, L"Quit", 1);
 
       if(auth_level != 3)
       {
@@ -2332,21 +2482,14 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
       {
         wchar_t buf[256];
 
-        if(showurl)
-        {
-          print_string(0, 320 + (selection == 4 ? bump : 0), 410, L"Log Out", 1);
-        }
-        else
-        {
-          print_string(0, 320 + (selection == 3 ? bump : 0), 370, L"Log Out", 1);
-        }
+        print_string(0, 320 + (selection == 3 ? bump : 0), 370, L"Log Out", 1);
 #ifndef WIN32
         swprintf(buf, sizeof(buf), L"Logged in as %ls.", username);
 #else
         swprintf(buf, L"Logged in as %ls.", username);
 #endif
 
-        print_string(1, 320, 440, buf, 1);
+        print_string(1, 320, 410, buf, 1);
       }
     }
     else if(mode == MODE_LOUNGE || mode == MODE_SCORE)
@@ -2357,22 +2500,27 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
       if(repeat_key && now > repeat_time)
       {
-        int length = 0;
-
         if(auth_level == 3 || mode == MODE_SCORE)
         {
-          while(length < 256 && message[length])
-            ++length;
-
           if(repeat_key == '\b')
           {
-            if(length)
-              message[length - 1] = 0;
+            message_backspace();
           }
-          else if(length < 256)
+          else if(repeat_key == '\v')
           {
-            message[length] = repeat_key;
-            message[length + 1] = 0;
+            message_delete();
+          }
+          else if(repeat_key == '\n')
+          {
+            cursor_left();
+          }
+          else if(repeat_key == '\r')
+          {
+            cursor_right();
+          }
+          else
+          {
+            message_insert(repeat_key);
           }
         }
 
@@ -2423,11 +2571,11 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
             print_string(1, 10, 129 + i * 17, privchatlog[(i + privchatlogpos) % CHAT_LINES], 0);
         }
 
-        size_t length = 0;
+        size_t length = message_length();
         size_t off = 0;
 
-        while(length < 256 && message[length])
-          ++length;
+        if (message_cursor > length)
+          message_cursor = length;
 
         if(string_width(1, message, length) > CHAT_WIDTH)
         {
@@ -2435,17 +2583,23 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
             if(string_width(1, message + off, length - off) < CHAT_WIDTH)
               break;
         }
-#ifndef WIN32
+        // Insert cursor (or space to make it blinking) :
+        memmove(message+message_cursor+1,
+                message+message_cursor,
+                sizeof(message[0])*(length-message_cursor+1));
         if(SDL_GetTicks() % 500 < 400)
-          swprintf(buf, sizeof(buf), L"%ls_", message + off);
+          message[message_cursor] = L'|';
         else
+          message[message_cursor] = L' ';
+#ifndef WIN32
           swprintf(buf, sizeof(buf), L"%ls", message + off);
 #else
-        if(SDL_GetTicks() % 500 < 400)
-          swprintf(buf, L"%ls_", message + off);
-        else
           swprintf(buf, L"%ls", message + off);
 #endif
+        // Remove inserted cursor :
+        memmove(message+message_cursor,
+                message+message_cursor+1,
+                sizeof(message[0])*(length-message_cursor+1));
 
         if(off != 0)
           print_string(1, 2, 455, L"..", 0);
@@ -2511,20 +2665,11 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
           }
           else if(auth_level == 3 || mode == MODE_SCORE)
           {
-            while(length < 256 && message[length])
-              ++length;
-
-            if(length < 256)
-            {
-              message[length] = event.key.keysym.unicode;
-              message[length + 1] = 0;
-
-              repeat_sym = event.key.keysym.sym;
-              repeat_key = event.key.keysym.unicode;
-              repeat_time = SDL_GetTicks() + 250;
+            message_insert(event.key.keysym.unicode);
+            repeat_sym = event.key.keysym.sym;
+            repeat_key = event.key.keysym.unicode;
+            repeat_time = SDL_GetTicks() + 250;
             }
-          }
-
           continue;
         }
 
@@ -2532,7 +2677,7 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
         {
         case SDLK_ESCAPE:
 
-          if(mode == MODE_MAIN_MENU)
+          if(mode == MODE_MAIN_MENU || mode == MODE_MODE_SELECT || mode == MODE_SINGLEPLAYER_MENU)
           {
             show_splash();
 
@@ -2666,8 +2811,7 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
               if(mode == MODE_SCORE)
               {
-                wchar_t buf[269];
-
+                wchar_t buf[270];
 #ifndef WIN32
                 swprintf(buf, sizeof(buf), L"You: %ls", message);
 #else
@@ -2694,11 +2838,59 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
             break;
           }
 
-          /* Fall through */
+          /* Fall through, used for shooting */
 
         case SDLK_SPACE:
 
-          if(mode == MODE_MAIN_MENU)
+          if(mode == MODE_MODE_SELECT)
+          {
+            if(selection == 0)
+            {
+              mode = MODE_SINGLEPLAYER_MENU;
+            }
+            else if(selection == 1)
+            {
+              selection = 0;
+
+              mode = MODE_MAIN_MENU;
+              time_stepms = 8;
+              time_step = 0.008f;
+
+              connect_to_master();
+            }
+            else if(selection == 2)
+            {
+              show_splash();
+
+              exit(EXIT_SUCCESS);
+            }
+          }
+          else if(mode == MODE_SINGLEPLAYER_MENU)
+          {
+            if(selection == 0)
+            {
+              saved_level = level;
+              play_single_player();
+
+              SDL_BlitSurface(logo, 0, screen, 0);
+            }
+            else if(selection == 1)
+            {
+              saved_level = level;
+              level = 255;
+              play_single_player();
+
+              SDL_BlitSurface(logo, 0, screen, 0);
+              selection = 0;
+            }
+            else if(selection == 2)
+            {
+              show_splash();
+
+              exit(EXIT_SUCCESS);
+            }
+          }
+          else if(mode == MODE_MAIN_MENU)
           {
             if(selection == 0)
             {
@@ -2715,22 +2907,13 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
               mode = MODE_LOUNGE;
             }
-            else if(selection == 2 && showurl)
-            {
-              if(fullscreen)
-                toggle_fullscreen();
-
-#ifdef WIN32
-              system("start http://www.junoplay.com/pengupop_buy_full");
-#endif
-            }
-            else if((selection == 2 && !showurl) || (selection == 3 && showurl))
+            else if(selection == 2)
             {
               show_splash();
 
               exit(EXIT_SUCCESS);
             }
-            else if((selection == 3 && !showurl) || (selection == 4 && showurl))
+            else if(selection == 3)
             {
               struct data_packet p;
               p.packet_type = packet_logout;
@@ -2742,13 +2925,13 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
               selection = 0;
             }
           }
-          else if(mode == MODE_GAME)
+          else if(mode == MODE_GAME && event.key.keysym.sym != SDLK_RETURN)
           {
-            if(!has_shot && !input_locked && (event.key.keysym.sym != SDLK_RETURN))
+            if(!has_shot && !input_locked)
             {
               update_movement = 1;
               has_shot = 1;
-              shoot(&players[0]);
+              shoot(&players[0], -1, -1);
               add_event(tickidx, EVENT_SHOOT | players[0].next_bubble);
             }
           }
@@ -2762,24 +2945,23 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
         case SDLK_UP:
 
-          if(mode == MODE_MAIN_MENU)
+          if(mode == MODE_MODE_SELECT || mode == MODE_SINGLEPLAYER_MENU)
           {
             last_tick = 0;
 
-            if(showurl)
-            {
-              if(selection == 0)
-                selection = (auth_level == 3) ? 4 : 3;
-              else
-                --selection;
-            }
+            if(selection == 0)
+              selection = 2;
             else
-            {
-              if(selection == 0)
-                selection = (auth_level == 3) ? 3 : 2;
-              else
-                --selection;
-            }
+              --selection;
+          }
+          else if(mode == MODE_MAIN_MENU)
+          {
+            last_tick = 0;
+
+            if(selection == 0)
+              selection = (auth_level == 3) ? 3 : 2;
+            else
+              --selection;
           }
           else if(mode == MODE_GAME)
           {
@@ -2787,7 +2969,7 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
             {
               update_movement = 1;
               has_shot = 1;
-              shoot(&players[0]);
+              shoot(&players[0], -1, -1);
               add_event(tickidx, EVENT_SHOOT | players[0].next_bubble);
             }
           }
@@ -2796,52 +2978,60 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
         case SDLK_DOWN:
 
-          if(mode == MODE_MAIN_MENU)
+          if(mode == MODE_MODE_SELECT || mode == MODE_SINGLEPLAYER_MENU)
           {
-            last_tick = SDL_GetTicks();
+            last_tick = 0;
 
-            if(showurl)
-            {
-              if(selection == ((auth_level == 3) ? 4 : 3))
-                selection = 0;
-              else
-                ++selection;
-            }
+            if(selection == 2)
+              selection = 0;
             else
-            {
-              if(selection == ((auth_level == 3) ? 3 : 2))
-                selection = 0;
-              else
-                ++selection;
-            }
+              ++selection;
+          }
+          else if(mode == MODE_MAIN_MENU)
+          {
+            last_tick = 0;
+
+            if(selection == ((auth_level == 3) ? 3 : 2))
+              selection = 0;
+            else
+              ++selection;
           }
 
           break;
 
         case SDLK_LEFT:
-
           if(mode == MODE_GAME)
           {
             update_movement = 1;
             players[0].right = -1;
             add_event(tickidx, EVENT_LEFT);
           }
-
+          else if(mode == MODE_LOUNGE || mode == MODE_SCORE)
+          {
+            cursor_left();
+            repeat_sym = event.key.keysym.sym;
+            repeat_key = '\n';
+            repeat_time = SDL_GetTicks() + 250;
+          }
           break;
 
         case SDLK_RIGHT:
-
           if(mode == MODE_GAME)
           {
             update_movement = 1;
             players[0].right = 1;
             add_event(tickidx, EVENT_RIGHT);
           }
-
+          else if(mode == MODE_LOUNGE || mode == MODE_SCORE)
+          {
+            cursor_right();
+            repeat_sym = event.key.keysym.sym;
+            repeat_key = '\r';
+            repeat_time = SDL_GetTicks() + 250;
+          }
           break;
 
         case SDLK_BACKSPACE:
-
           if(mode == MODE_LOUNGE || mode == MODE_SCORE)
           {
             int length = 0;
@@ -2864,18 +3054,37 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
             }
             else if(auth_level == 3 || mode == MODE_SCORE)
             {
-              while(length < 256 && message[length])
-                ++length;
-
-              if(length > 0)
-                message[--length] = 0;
+              message_backspace();
             }
 
             repeat_sym = event.key.keysym.sym;
             repeat_key = '\b';
             repeat_time = SDL_GetTicks() + 250;
           }
+          break;
 
+        case SDLK_DELETE:
+          if((mode == MODE_LOUNGE && auth_level == 3) || mode == MODE_SCORE)
+          {
+            message_delete();
+            repeat_sym = event.key.keysym.sym;
+            repeat_key = '\v';
+            repeat_time = SDL_GetTicks() + 250;
+          }
+          break;
+
+        case SDLK_HOME:
+          if((mode == MODE_LOUNGE && auth_level == 3) || mode == MODE_SCORE)
+          {
+             message_cursor = 0;
+          }
+          break;
+
+        case SDLK_END:
+          if((mode == MODE_LOUNGE && auth_level == 3) || mode == MODE_SCORE)
+          {
+            message_cursor = message_length();
+          }
           break;
 
         case 's':
@@ -2886,7 +3095,32 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
 
         case 'f':
 
-          toggle_fullscreen();
+#ifndef WIN32
+          SDL_WM_ToggleFullScreen(screen);
+#else
+          if(fullscreen)
+            screen = SDL_SetVideoMode(width, height, 0, SDL_SWSURFACE);
+          else
+            screen = SDL_SetVideoMode(width, height, 0, SDL_SWSURFACE | SDL_FULLSCREEN);
+
+          if(mode != MODE_GAME)
+          {
+            SDL_BlitSurface(logo, 0, screen, 0);
+          }
+          else
+          {
+            SDL_BlitSurface(background, 0, screen, 0);
+
+            for(i = 0; i < 2; ++i)
+            {
+              players[i].dirty_minx = 0;
+              players[i].dirty_miny = 0;
+              players[i].dirty_maxx = max_field_width * 32;
+              players[i].dirty_maxy = 440;
+            }
+          }
+#endif
+          fullscreen = !fullscreen;
 
           break;
 
@@ -2933,12 +3167,15 @@ int PASCAL WinMain(HINSTANCE instance, HINSTANCE previnstance,
       }
     }
 
-    if(players[0].evil_bubble_count && ((last_evil + (100 / time_stepms) < tickidx) || peer_version < 3) && !input_locked)
+    if(players[0].evil_bubble_count && ((last_evil + (100 / time_stepms) < tickidx) || peer_version < 3))
     {
       struct player_state* p = &players[0];
 
       i = 0;
 
+      if(p->evil_bubble_count > 4)
+        p->evil_bubble_count = 4;
+
       while(p->evil_bubble_count && i < sizeof(p->mbubbles) / sizeof(p->mbubbles[0]))
       {
         if(p->mbubbles[i].color != 0 || p->mbubbles[i].lastpaintx != INT_MIN)
